from datetime import datetime

from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class BinanceAPI(BlockchainAPI):
    """
    coins: binance coin
    API docs: https://docs.binance.org/api-reference/dex-api/paths.html
    Explorer: https://explorer.binance.org
    """

    active = True

    symbol = 'BNB'
    base_url = 'https://dex.binance.org/api/v1'
    rate_limit = 0
    coef = 1
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/account/{address}',
        'get_txs': '/transactions?address={address}&offset={offset}&limit={limit}'
    }

    @on_failure_return_none()
    def get_balance(self):
        response = self.request('get_balance', address=self.address)
        if not response:
            return None

        try:
            return [{
                'symbol': bal['symbol'],
                'amount': float(bal['free']) * self.coef
            } for bal in response['balances']]
        except ValueError:
            return None

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        response = self.request('get_txs',
                                address=self.address,
                                offset=offset,
                                limit=limit)

        return [self.parse_tx(t) for t in response['tx']]

    def parse_tx(self, tx):
        from_address = tx['fromAddr']
        to_address = tx['toAddr']
        amount = tx['value']
        fee = tx['txFee']
        txhash = tx['txHash']
        timestamp = tx['timeStamp']

        if from_address == self.address:
            direction = 'outgoing'
        else:
            direction = 'incoming'

        return {
            'date': datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%fZ'),
            'from_address': from_address,
            'to_address': to_address,
            'amount': amount * self.coef,
            'fee': fee * self.coef,
            'gas': {},
            'hash': txhash,
            'confirmed': None,
            'is_error': False,
            'type': 'normal',
            'kind': 'transaction',
            'direction': direction,
            'raw': tx
        }



from datetime import datetime
import pytz

from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class OntioAPI(BlockchainAPI):
    """
    coins: ontology
    API docs: https://dev-docs.ont.io/#/docs-en/API/02-restful_api
              https://github.com/ontio/ontology-explorer/tree/master/back-end-projects/Explorer/src/main/java/com/github/ontio/controller
    Explorer: https://explorer.ont.io
    """

    active = True

    symbol = 'ONT'
    base_url = 'https://explorer.ont.io'
    rate_limit = 0
    coef = 1
    max_items_per_page = 20
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/api/v1/explorer/address/{address}/0/1',
        'get_txs': '/api/v1/explorer/address/{address}/{limit}/{page}'
    }

    @on_failure_return_none()
    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None

        return [{"symbol": item['AssetName'].upper(), "amount": item['Balance']}
                for item in response['Result']['AssetBalance']]


    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        if limit is None:
            limit = self.max_items_per_page
        if offset is None:
            offset = 1

        response = self.request('get_txs',
                                address=self.address,
                                limit=limit,
                                page=offset)

        txs = response['Result']['TxnList']

        txs_result = []

        for tx in txs:
            for tx_transfer in tx['TransferList']:
                txs_result.append({
                    'date': datetime.fromtimestamp(tx['TxnTime'], pytz.utc),
                    'from_address': tx_transfer['FromAddress'],
                    'to_address': tx_transfer['ToAddress'],
                    'amount': tx_transfer['Amount'] * self.coef,
                    'fee': tx['Fee'] * self.coef,
                    'hash': tx['TxnHash'],
                    'confirmed': None,
                    'is_error': False,
                    'type': 'normal',
                    'kind': 'transaction',
                    'direction': 'outgoing' if tx_transfer['FromAddress'] == self.address else 'incoming',
                    'status': 'confirmed' if tx['ConfirmFlag'] == 1 else 'unconfirmed',
                    'raw': tx
                })


        return txs_result



from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class InsightAPI(BlockchainAPI):
    """
    coins: zcash, ravencoin, bitcoin
    """

    active = True

    rate_limit = 0
    coef = 1e-8
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/addr/{address}/balance',
    }

    @on_failure_return_none()
    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None

        retval = response * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]


class BitpayAPI(InsightAPI):
    """
    coins: bitcoin
    API docs: 
    Explorer: 
    """
    symbol = 'BTC'
    base_url = 'https://insight.bitpay.com/api'


class MercerweissAPI(InsightAPI):
    """
    coins: zcash
    API docs: http://insight.mercerweiss.com/api
    Explorer: http://insight.mercerweiss.com/
    """
    symbol = 'ZEC'
    base_url = 'http://insight.mercerweiss.com/api'


class RavencoinAPI(InsightAPI):
    """
    coins: ravencoin
    API docs: https://github.com/RavenDevKit/insight-api
    Explorer: https://ravencoin.network
    """
    symbol = 'RVN'
    base_url = 'https://ravencoin.network/api'


class InsightLitecoreAPI(InsightAPI):
    """
    coins: litecoin
    API docs: 
    Explorer: https://insight.litecore.io
    """
    active = False
    symbol = 'LTC'
    base_url = 'https://insight.litecore.io/api'



from datetime import datetime
import pytz
from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none,
    APIError
)


class AlethioAPI(BlockchainAPI):
    """
    Ethereum
    API docs: https://docs.aleth.io/api
    Explorer:
    notice: API key is needed
    """

    active = True

    symbol = 'ETH'
    base_url = 'https://api.aleth.io/v1'
    rate_limit = 0
    coef = 1e-18
    max_items_per_page = 100
    page_offset_step = None
    confirmed_num = None
    has_next = {'normal': True,
                'token': True}
    collect_logs = True

    supported_requests = {
        'get_balance': '/ether-balances?filter[account]={address}',
        'get_token_balances': '/token-balances?filter[account]={address}',
        'get_token_info': '/tokens/{token_id}',
        'get_txs':
            '/transactions?filter[account]={address}&page[limit]={limit}',
        'get_txs_next': None,
        'get_token_txs': '/token-transfers?filter[account]={address}'
                         '&page[limit]={limit}',
        'get_token_txs_next': None,
        'get_logs': None,
        'get_info': '/accounts/{address}'
    }

    def __init__(self, address, api_key=None):
        self.has_next['normal'] = True
        self.has_next['token'] = True
        self._info = None
        super().__init__(address, api_key)

    def _query_api(self, request_method, **kwargs):
        return self.request(request_method,
                            headers={
                                'Authorization': 'Bearer {}'.format(
                                    self.api_key)},
                            **kwargs)

    @on_failure_return_none()
    def get_balance(self):
        """
        Returns a list of all balances, both Ethereum and token ones

        :return: list
        """
        response = self._query_api('get_balance',
                                   address=self.address)
        if not response:
            return None

        retval_eth = int(response['data'][0]['attributes']['balance']) * \
            self.coef
        balance_eth = [{'symbol': self.symbol, 'amount': retval_eth}]
        balances_tokens = self._get_token_balances()
        if balances_tokens is None:
            return None

        return balance_eth + balances_tokens

    def _get_token_balances(self):
        response = self._query_api('get_token_balances',
                                   address=self.address)
        if not response:
            return None

        balances = []

        for token in response['data']:
            token_id = token['relationships']['token']['data']['id']
            token_info = self._query_api('get_token_info',
                                         token_id=token_id)
            if not token_info:
                return None

            token_info_attrs = token_info['data']['attributes']
            token_decimals = int(token_info_attrs['decimals'])
            token_symbol = token_info_attrs['symbol']
            token_name = token_info_attrs['name']
            bal = token['attributes']['balance']
            if bal is None:
                continue
            token_balance = int(bal) * \
                pow(10, -token_decimals)

            balances.append({'contract_address': token_id,
                             'name': token_name,
                             'symbol': token_symbol,
                             'amount': token_balance})

        return balances

    def get_txs(self, page=None, limit=None, unconfirmed=False,
                collect_logs=True):
        """
        Returns the list of collected transactions, subsequent calls
        download next transactions if available

        :param page: not used at all here
        :param limit: page limit; 100 is the maximum with AlethioAPI
        :param unconfirmed: not used at all here
        :param collect_logs: True if you'd like to collect eventlogs for each
                             transaction, otherwise False
        :return: list
        """
        self.collect_logs = collect_logs
        return self._get_txs(tx_type='normal', limit=limit)

    def get_token_txs(self, page=None, limit=None, unconfirmed=False):
        """
        Returns a list of collected token transactions, subsequent calls
        download next transactions if available

        :param page: not used at all here
        :param limit: page limit; 100 is the maximum with AlethioAPI
        :param unconfirmed: not used here at all
        :return: list
        """
        return self._get_txs(tx_type='token', limit=limit)

    def _get_txs(self, tx_type='normal', page=None, limit=None,
                 unconfirmed=False):
        if tx_type == 'normal':
            fetch_req = 'get_txs'
            fetch_next_req = 'get_txs_next'
            parser = self._parse_tx
        else:
            fetch_req = 'get_token_txs'
            fetch_next_req = 'get_token_txs_next'
            parser = self._parse_token_tx

        if not self.has_next[tx_type]:
            return []

        if limit is None:
            limit = self.max_items_per_page

        if limit > self.max_items_per_page:
            limit = self.max_items_per_page

        if self.supported_requests[fetch_next_req] is None:
            txs = self._query_api(fetch_req, address=self.address,
                                  limit=limit)
        else:
            txs = self._query_api(fetch_next_req)

        if txs['meta']['page']['hasNext']:
            self.has_next[tx_type] = True
            self.supported_requests[fetch_next_req] = \
                txs['links']['next'].replace(self.base_url, '')
        else:
            self.has_next[tx_type] = False

        parsed_txs = [parser(tx) for tx in txs['data']]
        return parsed_txs

    def _parse_token_tx(self, tx):
        attributes = tx['attributes']
        relationships = tx['relationships']

        tx_symbol = attributes['symbol']
        tx_from_address = relationships['from']['data']['id']
        tx_to_address = relationships['to']['data']['id']
        tx_contract_address = relationships['token']['data']['id']

        tx_direction = self._get_tx_direction(tx_from_address, tx_to_address,
                                              tx_contract_address)
        tx_token_data = {
            'name': None,  # there's no such field
            'symbol': tx_symbol,
            'decimals': attributes['decimals']
        }

        return {
            'symbol': tx_symbol,
            'date':
                datetime.fromtimestamp(int(attributes['blockCreationTime']),
                                       pytz.utc),
            'from_address': tx_from_address,
            'to_address': tx_to_address,
            'contract_address': tx_contract_address,
            'amount': float(attributes['value']) *
                pow(10, -attributes['decimals']),
            'fee': None,
            'gas': {
                'gas': float(attributes['transactionGasLimit']),
                'gas_price': float(attributes['transactionGasPrice']),
                'cumulative_gas_used': None,
                'gas_used': float(attributes['transactionGasUsed'])},
            'hash': relationships['transaction']['data']['id'],
            'confirmations': None,
            'confirmed': None,
            'is_error': False,  # should be always valid
            'type': 'token',
            'kind': 'transaction',
            'direction': tx_direction,
            'token_data': tx_token_data,
            'raw': tx
        }

    def _parse_tx(self, tx):
        attributes = tx['attributes']
        relationships = tx['relationships']

        tx_from_address = relationships['from']['data']['id']
        tx_to_address = relationships['to']['data']['id']
        if attributes['msgType'] in ('CallTx', 'CreateTx'):
            tx_contract_address = tx_to_address
        else:
            tx_contract_address = ''

        tx_direction = self._get_tx_direction(tx_from_address, tx_to_address,
                                              tx_contract_address)
        log_entries_req = relationships['logEntries']['links']['related']
        self.supported_requests['get_logs'] = \
            log_entries_req.replace(self.base_url, '')

        if self.collect_logs:
            parsed_logs = self._get_logs()
        else:
            parsed_logs = None

        return {
            'symbol': 'ETH',
            'date':
                datetime.fromtimestamp(int(attributes['blockCreationTime']),
                                       pytz.utc),
            'from_address': tx_from_address,
            'to_address': tx_to_address,
            'contract_address': tx_contract_address,
            'amount': float(attributes['value']) * self.coef,
            'fee': float(attributes['fee']) * self.coef,
            'gas': {
                'gas': float(attributes['msgGasLimit']),
                'gas_price': float(attributes['txGasPrice']),
                'cumulative_gas_used': None,
                'gas_used': float(attributes['txGasUsed'])},
            'hash': attributes['txHash'],
            'confirmations': None,
            'confirmed': None,
            'is_error': attributes['msgError'],
            'type': 'normal',
            'kind': 'transaction',
            'direction': tx_direction,
            'token_data': None,
            'raw': tx,
            'event_logs': parsed_logs
        }

    def _get_tx_direction(self, tx_from_address, tx_to_address,
                          tx_contract_address):
        tx_direction = None
        if self.address.lower() == tx_from_address.lower():
            tx_direction = 'outgoing'
        elif self.address.lower() == tx_to_address.lower():
            tx_direction = 'incoming'
        elif not tx_contract_address:
            tx_direction = 'outgoing'

        return tx_direction

    def _get_logs(self):
        response = self._query_api('get_logs')
        if not response:
            return None

        logs = []

        next_logs = True
        while next_logs:
            for log in response['data']:
                logs.append(self._parse_log(log))
            next_logs = response['meta']['page']['hasNext']

        return logs

    def _parse_log(self, log):
        attributes = log['attributes']
        relationships = log['relationships']

        if attributes['eventDecodedError'] == '':
            event_name = attributes['eventDecoded']['event']
            inputs = []
            for inp in attributes['eventDecoded']['inputs']:
                if 'indexed' in inp:
                    is_indexed = inp['indexed']
                else:
                    is_indexed = False
                inputs.append({
                    'name': inp['name'],
                    'type': inp['type'],
                    'value': inp['value'],
                    'is_indexed': is_indexed
                })
        else:
            event_name = ''
            inputs = None

        result = {
            'log_data': attributes['logData'],
            'event': event_name,
            'inputs': inputs,
            'address': relationships['loggedBy']['data']['id']
        }

        topics = attributes['hasLogTopics']

        i = 0
        for topic in topics:
            result['topic{}'.format(i)] = topic
            i += 1

        return result

    @property
    def info(self):
        if self._info is None:
            result = self._query_api('get_info', address=self.address)
            if result['data']['relationships']['contract']['data'] is None:
                self._info = 'address'
            else:
                # contract or token
                try:
                    self._query_api('get_token_info', token_id=self.address)
                    self._info = 'token'
                except APIError:
                    self._info = 'contract'

        return self._info



import dateutil.parser
from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class BlockchainosAPI(BlockchainAPI):
    """
    coins: boscoin
    API docs: http://devteam.blockchainos.org/docs/api/
    Explorer: https://explorer.boscoin.io/
    """

    active = True

    symbol = 'BOS'
    base_url = 'https://mainnet.blockchainos.org'
    rate_limit = 0
    coef = 1e-7
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/api/v1/accounts/{address}',
        'get_txs': '/api/v1/accounts/{address}/transactions?limit={limit}&reverse=true',
    }

    @on_failure_return_none()
    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None

        try:
            balance = int(response.get('balance'))
        except (KeyError,ValueError):
            return None

        return [{'symbol': self.symbol, 'amount': balance * self.coef}]

    def get_txs(self, limit=None):
        if not 'get_txs_next' in self.supported_requests:
            response = self.request('get_txs',
                                    address=self.address,
                                    limit=limit)
            self.supported_requests['get_txs_next'] = response['_links']['prev']['href']

        else:
            response = self.request('get_txs_next')
            self.supported_requests['get_txs_next'] = response['_links']['prev']['href']

        return [self.parse_tx(tx) for tx in response['_embedded']['records']]

    def parse_tx(self,tx):
        operations = []

        operation_url = tx['_links']['operations']['href']
        self.supported_requests['get_operations'] = operation_url.replace('{?cursor,limit,order}','')
        response = self.request('get_operations')

        for operation in response['_embedded']['records']:
            op_from_address = operation['source']
            op_to_address = operation['target']
            op_hash = operation['tx_hash']
            op_amount = operation['body']['amount']
            op_type = operation['type']
            op_confirmed = dateutil.parser.parse(operation['confirmed'])
            operations.append({'from_address': op_from_address,
                               'to_address': op_to_address,
                               'hash': op_hash,
                               'amount': float(op_amount) * self.coef,
                               'type': op_type,
                               'direction': 'outgoing' if op_from_address == self.address else 'incoming',
                               'confirmed': op_confirmed,
                               'raw': operation})


        return {
            'date': dateutil.parser.parse(tx['created']),
            'fee': float(tx['fee']) * self.coef,
            'hash': tx['hash'],
            'is_error': False,
            'kind': 'transaction',
            'status': 'confirmed',
            'operations': operations,
            'raw': tx
        }



from blockapi.services import (
    BlockchainAPI,
    APIKeyMissing,
    on_failure_return_none
)


class CryptoIDAPI(BlockchainAPI):
    """
    Multi coins
    API docs: https://chainz.cryptoid.info/api.dws
    Explorer: https://chainz.cryptoid.info
    """

    active = True

    symbol = None
    base_url = 'http://chainz.cryptoid.info'
    rate_limit = 0
    coef = 1e-8
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/{symbol}/api.dws?q=getbalance&a={address}&key={api_key}',
    }

    def __init__(self, address, api_key=None):
        if api_key is None:
            raise APIKeyMissing

        super().__init__(address, api_key)

    @on_failure_return_none()
    def get_balance(self):
        response = self.request('get_balance',
                                api_key=self.api_key,
                                symbol=self.symbol.lower(),
                                address=self.address)

        return [{'symbol': self.symbol, 'amount': response * self.coef}]


class CryptoIDLitecoinAPI(CryptoIDAPI):
    symbol = 'LTC'


class CryptoIDDashAPI(CryptoIDAPI):
    symbol = 'DASH'
    coef = 1


class CryptoIDGroestlcoinAPI(CryptoIDAPI):
    symbol = 'GRS'
    coef = 1



from datetime import datetime

import pytz

from blockapi.services import (
    BlockchainAPI,
    set_default_args_values,
    AddressNotExist,
    on_failure_return_none
)


class BlockchainInfoAPI(BlockchainAPI):
    """
    Bitcoin
    API docs: https://www.blockchain.com/api/blockchain_api
    Explorer: https://blockchain.info
    """

    active = True

    symbol = 'BTC'
    base_url = 'https://blockchain.info'
    rate_limit = 0
    coef = 1e-8
    max_items_per_page = 100
    page_offset_step = max_items_per_page
    confirmed_num = 6

    supported_requests = {
        'get_balance': '/balance?active={address}',
        'get_txs': '/multiaddr?active={address}&n={number}&offset={offset}'
    }

    def process_error_response(self, response):
        if response.text == 'Invalid Bitcoin Address':
            raise AddressNotExist()
        # else
        super().process_error_response(response)

    @on_failure_return_none()
    def get_balance(self):
        balance = self.request('get_balance', address=self.address)
        if not balance:
            return None
        retval = list(balance.values())[0]['final_balance'] * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]

    @set_default_args_values
    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        # always returns confimed transactions
        response = self.request(
            'get_txs',
            address=self.address,
            number=limit,
            offset=offset
        )
        return [self.parse_tx(t) for t in response['txs']]

    def parse_tx(self, tx):
        out_addresses = [o['addr'] for o in tx['out']]
        in_addresses = [i['prev_out']['addr'] for i in tx['inputs']]
        is_incoming = next((True for o in tx['out'] if o.get('xpub')), False)
        is_outgoing = next((True for i in tx['inputs'] if i['prev_out'].get('xpub')), False)
        direction = 'incoming' if is_incoming else 'outgoing'

        if is_outgoing:
            from_addresses = in_addresses
            to_addresses = out_addresses
        else:
            from_addresses = out_addresses
            to_addresses = in_addresses

        return {
            'date': datetime.fromtimestamp(tx['time'], pytz.utc),
            'from_address': from_addresses,
            'to_address': to_addresses, # multiple, TODO check it
            'amount': tx['balance'] * self.coef,
            'fee': tx['fee'] * self.coef,
            'hash': tx['hash'],
            'confirmed': None,
            'is_error': False,
            'type': 'normal',
            'kind': 'transaction',
            'direction': direction,
            'status': 'confirmed',
            'raw': tx
        }




import itertools
import unittest

from cfscrape import CloudflareCaptchaError

import blockapi
import blockapi.api
from blockapi.services import AddressNotExist, APIError, BadGateway, \
    GatewayTimeOut, InternalServerError, APIKeyMissing

test_addresses = {
    'BTC': [
        '1NuXUAnkWBYF3Fs9CkjfARYMacrVtoCrAM',
        #'ypub6WjHjrJLKSg8oQw1E4LGvQDJ2uofgMfJKLnv5Ha4NPRW4rf7LPXffMJ8EReixY1mUCc33SsiDUodUCTvCktFeN7ZW28GVhBXhNnoKYUqXbP',
        'xpub6CUGRUonZSQ4TWtTMmzXdrXDtypWKiKrhko4egpiMZbpiaQL2jkwSB1icqYh2cfDfVxdx4df189oLKnC5fSwqPfgyP3hooxujYzAu3fDVmz'
    ],
    'BCH': [
        '35hK24tcLEWcgNA4JxpvbkNkoAcDGqQPsP',
    ],
    'ADA': [
        'DdzFFzCqrhsdyjuJWYcfEWSuk4TB25HGv66bQKeU3KpvA78xfDbgu1rQdhaLHb5XvqPU7gfKS5mBoGmRNNhTSQ3H7iL1so8RygHMdNiN'
    ],
    'EOS': [],
    'ETH': [
        '0x1d0DcC8d8BcaFa8e8502BEaEeF6CBD49d3AFFCDC'
    ],
    'LTC': [
        # 'MBuTKxJaHMN3UsRxQqpGRPdA7sCfE1UF7n',
        '3QY7aJKtRHDy3a8V5RS99A813hie9YqjhF'
    ],
    'XTZ': [
        'tz1bDXD6nNSrebqmAnnKKwnX1QdePSMCj4MX'
    ],
    'DCR': [
        'DsXt3he1A9KB2uL1g3MJvbAbXEB1CxN2rNF'
        # current api doesn't support xpub key
        # 'dpubZFwoKEEJYVDxGo8bf2E4qwh6Qve9cku5gvaS5kC96hUdMT7SF9nymaLFeEFQaHy8a3SuiUJRL87rz3bfwFSFqErYVeHUg3xnzPjHftiofFu',
    ],
    'ATOM': [
        'cosmos1gn326f6sza44xt5kxrsdrnapp2sxhav03rhcsz'
    ],
    'NEO': [
        'AZnTM3mYbx9yzg8tb6hr7w9pAKntDmrtqk'
    ],
    'DOGE': [
        'DH5yaieqoZN36fDVciNyRueRGvGLR3mr7L'
    ],
    'ZEC': [
        't1fLdzB7mHQgdb3tD474V9wYtuknPfQSX5e'
    ],
    'DASH': [
        'XtAG1982HcYJVibHxRZrBmdzL5YTzj4cA1'
    ],
    'ETC': [
        '0x9d2BFC36106F038250c01801685785b16C86C60D'
    ],
    'ZEN': [
        'znZTLu1asaLWxB7EBqBRQ6DCnNyctYA3Rm4'
    ],
    'BNB': [
        'bnb1jxfh2g85q3v0tdq56fnevx6xcxtcnhtsmcu64m'
    ],
    'XLM': [
        'GDD7ABRF7BCK76W33RXDQG5Q3WXVSQYVLGEMXSOWRGZ6Z3G3M2EM2TCP'
    ],
    'RVN': [
        'RMJVXz8pn7EsBRqXYvDXU7UV7WrbGiDiHk'
    ],
    'TRX': [
        'TNNc1HGDUrRkowQxdcUaWyBodZXshuVtBp'
    ],
    'GRS': [
        'Fr5m2irs9vNWSAFXJK6KPtxqW9YWg384FX'
    ]
}

test_invalid_addresses = {
    'BTC': ['xxxx', ],
    'BCH': ['xxxx', ],
    'ADA': ['xxxx', ],
    'EOS': [],
    'ETH': ['xxxx', ],
    'LTC': ['xxxx', ],
    'XTZ': ['xxxx', ],
    'DCR': ['xxxx', ],
    'ATOM': ['xxxx', ],
    'NEO': ['xxxx', ],
    'DOGE': ['xxxx', ],
    'ZEC': ['xxxx', ],
    'DASH': ['xxxx', ],
    'ETC': ['xxxx', ],
    'ZEN': ['xxxx', ],
    'BNB': ['xxxx', ],
    'XLM': ['xxxx', ],
    'RVN': ['xxxx', ],
    'TRX': ['xxxx', ],
    'GRS': ['xxxx', ]
}


class BlockapiTestCase(unittest.TestCase):
    currencies = test_addresses.keys()

    def test_valid_address(self):
        for symbol in self.currencies:
            with self.subTest(symbol):
                api_classes = blockapi.get_api_classes_for_coin(symbol)
                addresses = test_addresses[symbol]

                for api_class, address in itertools.product(api_classes, addresses):
                    #if 'api_key' in api_class.__init__.__code__.co_varnames:
                    #    continue

                    try:
                        api_inst = api_class(address)
                    except APIKeyMissing:
                        # skip test if api key is required
                        continue

                    if not api_inst.xpub_support and address.startswith('xpub'):
                        # skip testing against xpub address if xpub is not supported
                        continue

                    try:
                        b = api_inst.get_balance()
                    except (AddressNotExist, BadGateway, GatewayTimeOut, InternalServerError, APIError, CloudflareCaptchaError):
                        self.fail("get_balance for {} [{}] failed unexpectedly with a valid address {}".format(
                            api_inst.__class__.__name__, symbol, address))

    def test_invalid_address(self):
        for symbol in self.currencies:
            with self.subTest(symbol):
                api_classes = blockapi.get_api_classes_for_coin(symbol)
                addresses = test_invalid_addresses[symbol]

                for api_class, address in itertools.product(api_classes, addresses):
                    #if 'api_key' in api_class.__init__.__code__.co_varnames:
                    #    continue

                    try:
                        api_inst = api_class(address)
                        assert not api_inst.address_info.valid
                    except APIKeyMissing:
                        continue

                    # try:
                    #     api_inst = api_class(address)
                    # except ValueError:
                    #     pass
                    # except APIKeyMissing:
                    #     continue
                    # else:
                    #     with self.assertRaises(
                    #             (blockapi.services.AddressNotExist, blockapi.services.APIError),
                    #             msg="API/currency: {}/{}".format(api_inst.__class__.__name__,
                    #                                              symbol)):
                    #         b = api_inst.get_balance()

    def test_get_balance(self):
        for symbol in self.currencies:
            with self.subTest(symbol):
                api_classes = blockapi.get_api_classes_for_coin(symbol)
                addresses = test_addresses[symbol]

                for api_class, address in itertools.product(api_classes, addresses):
                    #if 'api_key' in api_class.__init__.__code__.co_varnames:
                    #    continue

                    try:
                        api_inst = api_class(address)
                    except APIKeyMissing:
                        continue

                    if not api_inst.xpub_support and address.startswith('xpub'):
                        # skip testing against xpub address if xpub is not supported
                        continue


                    try:
                        b = api_inst.get_balance()
                    except:
                        self.fail("get_balance for {} [{}] - {} failed unexpectedly".format(api_inst.__class__.__name__,
                                                                                       symbol,address))
                    is_ok = True

                    try:
                        tmp = float(b)
                    except (ValueError, TypeError):
                        is_ok = False

                    if not is_ok:
                        is_ok = True
                        if not isinstance(b, list):
                            is_ok = False

                    if not is_ok:
                        self.fail("get_balance for {} [{}] failed unexpectedly - returned value is not a number or list".format(
                            api_inst.__class__.__name__, symbol))


                    # if symbol == 'cosmos':
                    #    # get both incoming and outgoing transactions
                    #    it = api_inst.get_incoming_txs()
                    #    ot = api_inst.get_outgoing_txs()
                    # else:
                    #    # get transactions
                    #    t = api_inst.get_txs()

    def test_random_balance(self):
        for symbol in self.currencies:
            with self.subTest(symbol):
                addresses = test_addresses[symbol]
                for address in addresses:
                    try:
                        blockapi.get_balance_from_random_api(symbol, address)
                    except:
                        self.fail("get_balance_from_random_api for {} failed unexpectedly".format(symbol))

    def test_rate_limits(self):
        for symbol in self.currencies:
            with self.subTest(symbol):
                addresses = test_addresses[symbol]
                for address in addresses:
                    for _ in range(2):
                        try:
                            blockapi.get_balance_from_random_api(symbol, address)
                        except _:
                            self.fail("repeated call for {} failed unexpectedly".format(symbol))


# TODO check obligatory fields in response
def check_obligatory_fields(method, args, kwargs, obligatory_fields):
    response = method(*args, **kwargs)
    if not response:
        return True

    response_keys = []
    if type(response) == dict:
        response_keys = response.keys
    elif type(response) == list:
        response_keys = response[0].keys

    missing_fields = []
    for obl_field in obligatory_fields:
        if obl_field not in response_keys:
            missing_fields.append(obl_field)


if __name__ == "__main__":
    unittest.main()



import dateutil.parser

from blockapi.services import (
    BlockchainAPI,
    set_default_args_values,
    APIError,
    on_failure_return_none
)


class TzscanAPI(BlockchainAPI):
    """
    Tezos
    API docs: https://tzscan.io/api
    Explorer: https://tzscan.io
    """

    symbol = 'XTZ'
    base_url = None # endpoint is created in runtime
    rate_limit = 0
    coef = 1e-6
    max_items_per_page = 50
    page_offset_step = 1

    supported_requests = {
        'get_balance': '/v3/balance_from_balance_updates/{address}',
        'get_operations': '/v3/operations/{address}?type={type}&p={page_offset}&number={number}',
        'get_rewards': '/v3/rewards_split_cycles/{address}?p={page_offset}&number={number}',
        'get_rewards_split': '/v2/rewards_split/{address}?cycle={cycle}&number=50&p={page}',
        'get_bakings': '/v3/cycle_bakings/{address}?p={page_offset}&number={number}',
        'get_endorsements': '/v3/cycle_endorsements/{address}?p={page_offset}&number={number}'
    }

    def __init__(self, address, api_key=None):
        super().__init__(address)
        # this API is very unstable, add some functionality to prevent fails
        self._base_url_temp = 'https://api{num}.dunscan.io'  # num = 1-6
        self._api_nums = [1, 2, 3, 4, 5, 6]

    @on_failure_return_none()
    def get_balance(self):
        balance = self._safe_request('get_balance', address=self.address)
        return [{'symbol': self.symbol,
                 'amount': float(balance['spendable']) * self.coef}]

    def get_rewards(self, offset=0, limit=50):
        rewards = self._safe_request('get_rewards', address=self.address,
            page_offset=offset, number=limit)

        return [self.get_reward_details(reward, int(reward['cycle'])) for reward in rewards]

    def get_reward_details(self, reward_v3, cycle=None):
        reward = self._safe_request(
            'get_rewards_split', address=self.address, cycle=cycle,page=0)

        total_staking_balance = float(reward['delegate_staking_balance'])
        baker_balance = total_staking_balance
        total_rewards = float(reward['blocks_rewards']) + \
                        float(reward['endorsements_rewards']) + \
                        float(reward['fees']) + \
                        float(reward['future_blocks_rewards']) + \
                        float(reward['future_endorsements_rewards']) + \
                        float(reward['gain_from_denounciation_baking']) - \
                        float(reward['lost_deposit_from_denounciation_baking']) - \
                        float(reward['lost_fees_denounciation_baking']) - \
                        float(reward['lost_rewards_denounciation_baking']) + \
                        float(reward['gain_from_denounciation_endorsement']) - \
                        float(reward['lost_deposit_from_denounciation_endorsement']) - \
                        float(reward['lost_fees_denounciation_endorsement']) - \
                        float(reward['lost_rewards_denounciation_endorsement']) + \
                        float(reward['revelation_rewards']) - \
                        float(reward['lost_revelation_rewards']) - \
                        float(reward['lost_revelation_fees'])

        total_staking_balance = round(float(total_staking_balance) / 1000000, 6)


        total_delegators = int(reward['delegators_nb'])
        pages = total_delegators / 50

        delegations = []
        delegators = 0
        page = 0

        while True:
            for del_balance in reward['delegators_balance']:
                delegator_address = del_balance[0]['tz']
                delegator_balance = int(del_balance[1]) * self.coef
                delegations.append({'address': delegator_address,
                                    'balance': delegator_balance,
                                    'share': (float(delegator_balance) / total_staking_balance) * 100,
                                    'rewards_share': (delegator_balance/total_staking_balance) * (total_rewards * self.coef)})
                delegators += 1

            page += 1
            if page < pages:
                reward = self._safe_request(
                    'get_rewards_split', address=self.address, cycle=cycle,page=page)
            else:
                break

        return { 'staking_balance': total_staking_balance, 
                 'cycle': cycle,
                 'status': reward_v3['status']['status'],
                 'delegators': delegators,
                 'delegations': delegations,
                 'end_rewards': float(reward['endorsements_rewards']) * self.coef,
                 'blocks_rewards': float(reward['blocks_rewards']) * self.coef,
                 'extra_rewards': float(reward['revelation_rewards']) * self.coef,
                 'losses': float(reward['lost_revelation_rewards']) * self.coef,
                 'fees': float(reward['fees']) * self.coef
        }

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        return self._get_operations(
            'Transaction', self.parse_tx, offset, limit)

    def get_activations(self, offset=None, limit=None, unconfirmed=False):
        return self._get_operations(
            'Activation', self.parse_activation, offset, limit)

    def get_originations(self, offset=None, limit=None, unconfirmed=False):
        return self._get_operations(
            'Origination', self.parse_origination, offset, limit)

    def get_delegations(self, offset=None, limit=None, unconfirmed=False):
        return self._get_operations(
            'Delegation', self.parse_delegation, offset, limit)

    @set_default_args_values
    def _get_operations(self, op_type, parse, offset=None, limit=None):
        """Get all operations by type
        @op_type in [Transaction, Origination, Delegation, Activation, ...]"""

        operations = self._safe_request(
            'get_operations',
            address=self.address,
            type=op_type,
            page_offset=offset,
            number=limit
        )

        parsed_txs = []
        for tx in operations:
            parsed_txs += parse(tx)
        return parsed_txs

    def parse_tx(self, tx):
        # tezos can have multiple txs in single tx (with one hash)
        parsed = []
        for op in tx.get('type', {}).get('operations', []):
            parsed.append({
                'date': dateutil.parser.parse(op['timestamp']),
                'from_address': op['src']['tz'],
                'to_address': op['destination']['tz'],
                'amount': float(op['amount']) * self.coef,
                'fee': None if op['fee'] == -1 else op['fee'] * self.coef,
                'gas_limit': None if int(op['gas_limit']) == -1 else int(op['gas_limit']),
                'hash': tx['hash'],
                'confirmed': None,
                'is_error': op['failed'],
                'type': 'internal' if op['internal'] else 'normal',
                'kind': op['kind'].lower(),
                'direction': 'outgoing' if self.address == op['src']['tz'] else 'incoming',
                'raw': tx
            })
            return parsed

    def parse_delegation(self, tx):
        parsed = []
        for op in tx.get('type', {}).get('operations', []):
            parsed.append({
                'date': dateutil.parser.parse(op['timestamp']),
                'source_address': op['src']['tz'],
                'delegate': op['delegate']['tz'],
                'fee': None if op['fee'] == -1 else op['fee'] * self.coef,
                'gas_limit': None if int(op['gas_limit']) == -1 else int(op['gas_limit']),
                'hash': tx['hash'],
                'is_error': op['failed'],
                'type': 'internal' if op['internal'] else 'normal',
                'kind': op['kind'].lower(),
                'raw': tx
            })
            return parsed

    def parse_activation(self, tx):
        parsed = []
        for op in tx.get('type', {}).get('operations', []):
            parsed.append({
                'date': dateutil.parser.parse(op['timestamp']),
                'secret': op['secret'],
                'balance': op['balance'],
                'hash': tx['hash'],
                'kind': op['kind'].lower(),
                'raw': tx
            })
            return parsed

    def parse_origination(self, tx):
        parsed = []
        for op in tx.get('type', {}).get('operations', []):
            parsed.append({
                'date': dateutil.parser.parse(op['timestamp']),
                'originator': op['src']['tz'],
                'manager': op['managerPubkey']['tz'],
                'balance': op['balance'],
                'spendable': op['spendable'],
                'delegatable': op['delegatable'],
                'delegate': op['delegate']['tz'],
                'delegate_alias': op['delegate']['alias'],
                'burnt': op['burn_tez'] * self.coef,
                'fee': None if op['fee'] == -1 else op['fee'] * self.coef,
                'gas_limit': None if int(op['gas_limit']) == -1 else int(op['gas_limit']),
                'hash': tx['hash'],
                'is_error': op['failed'],
                'type': 'internal' if op['internal'] else 'normal',
                'kind': op['kind'].lower(),
                'raw': tx
            })
            return parsed

    @set_default_args_values
    def get_endorsements(self, offset=None, limit=None):
        ends = self._safe_request(
            'get_endorsements',
            address=self.address,
            page_offset=offset,
            number=limit
        )
        return [self.parse_endorsement(e) for e in ends]

    def parse_endorsement(self, e):
        return {
            'cycle': int(e['cycle']),
            'depth': int(e['depth']),
            'num_slots': {
                'all': int(e['slots']['count_all']),
                'miss': int(e['slots']['count_miss']),
                'steal': int(e['slots']['count_steal'])
            },
            'fee': int(e['tez']['fee']) * self.coef,
            'reward': int(e['tez']['reward']) * self.coef,
            'deposit': int(e['tez']['deposit']) * self.coef,
            'priority': float(e['priority'])
        }

    @set_default_args_values
    def get_bakings(self, offset=None, limit=None):
        baks = self._safe_request(
            'get_bakings',
            address=self.address,
            page_offset=offset,
            number=limit
        )
        return [self.parse_baking(b) for b in baks]

    def parse_baking(self, b):
        return {
            'cycle': int(b['cycle']),
            'depth': int(b['depth']),
            'num_blocks': {
                'all': int(b['count']['count_all']),
                'miss': int(b['count']['count_miss']),
                'steal': int(b['count']['count_steal'])
            },
            'fee': int(b['tez']['fee']) * self.coef,
            'reward': int(b['tez']['reward']) * self.coef,
            'deposit': int(b['tez']['deposit']) * self.coef,
            'priority': float(b['priority']),
            'bake_time': int(b['bake_time'])
        }

    def _safe_request(self, method, **params):
        api_nums = self._api_nums.copy()
        while api_nums:
            try:
                api_num = api_nums.pop(0)
                self.base_url = self._base_url_temp.format(num=api_num)
                response = self.request(method, **params)
            except APIError as e:
                ex = e
                continue
            else:
                return response

        # when all endpoints are checked and none of them is valid, raise exc
        raise ex



from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class BlockscoutAPI(BlockchainAPI):
    """
    coins: etc
    API docs: https://blockscout.com/etc/mainnet/api_docs
    Explorer: https://blockscout.com
    """

    active = True

    symbol = 'ETC'
    base_url = 'https://blockscout.com/etc/mainnet/api?'
    rate_limit = 0
    coef = 1e-18
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': 'module=account&action=eth_get_balance&address={address}',
    }

    @on_failure_return_none()
    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None

        retval = int(response['result'], 16) * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]



from datetime import datetime

import pytz

from blockapi.services import (
    BlockchainAPI,
    set_default_args_values,
    on_failure_return_none
)


class EtherscanAPI(BlockchainAPI):
    """
    Ethereum
    API docs: https://etherscan.io/apis
    Explorer: https://etherscan.io
    """

    symbol = 'ETH'
    base_url = 'https://api.etherscan.io'
    rate_limit = 2.5    # orignal rate_limit was 0.2  (5 req. / sec)
    coef = 1e-18
    start_offset = 1
    max_items_per_page = 10000
    page_offset_step = 1

    supported_requests = {
        'get_balance': '/api?module=account&action=balance&address={address}&tag=latest&api_key={api_key}',
        'get_txs': '/api?module=account&action={action}&offset={offset}&sort={sort}&page={page}&address={address}'
                   '&api_key={api_key}',
        'get_abi': '/api?module=contract&action=getabi&address={address}'
    }

    @on_failure_return_none()
    def get_balance(self):
        balance_dict = self.request(
            'get_balance',
            address=self.address,
            api_key=self.api_key
        )

        # returns only balance for ETH; ERC20 and ERC721 tokens are omitted
        if 'result' in balance_dict:
            try:
                retval = int(balance_dict['result']) * self.coef
            except ValueError:
                return None
        else:
            return None

        return [{'symbol': self.symbol, 'amount': retval}]

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        txs = self._get_txs('txlist', offset, limit)
        return [self.parse_tx(t, 'normal') for t in txs]

    def get_internal_txs(self, offset=None, limit=None, unconfirmed=False):
        txs = self._get_txs('txlistinternal', offset, limit)
        return [self.parse_tx(t, 'internal') for t in txs]

    def get_token_txs(self, offset=None, limit=None, unconfirmed=False):
        txs = self._get_txs('tokentx', offset, limit)
        return [self.parse_tx(t, 'token') for t in txs]

    def get_abi(self,contract):
        abi = self.request(
            'get_abi',
            address=contract)

        return abi

    @set_default_args_values
    def _get_txs(self, action, offset=None, limit=None):
        response = self.request(
            'get_txs',
            action=action,
            offset=limit,
            sort='desc',
            page=offset,
            address=self.address,
            api_key=self.api_key
        )

        if not response:
            return []
        result = response.get('result', [])
        if result is None:
            result = []

        status = response.get('status', '0')
        if status == '0':
            print('EtherscanAPI error: {}'.format(result))
            return []

        return result

    def parse_tx(self, tx, tx_type):
        direction = None
        if self.address.lower() == tx['from'].lower():
            direction = 'outgoing'
        elif self.address.lower() == tx['to'].lower():
            direction = 'incoming'
        elif not tx['contractAddress']:
            direction = 'outgoing'

        token_data = None
        if tx_type is 'token':
            token_data = {
                'name': tx.get('tokenName'),
                'symbol': tx.get('tokenSymbol'),
                'decimals': float(tx.get('tokenDecimal'))
            }
            symbol = token_data['symbol']
            amount = float(tx['value']) * pow(10, -token_data['decimals'])
        else:
            symbol = self.symbol
            amount = float(tx['value']) * self.coef

        return {
            'symbol': symbol,
            'date': datetime.fromtimestamp(int(tx['timeStamp']), pytz.utc),
            'from_address': tx['from'],
            'to_address': tx['to'],
            'contract_address': tx['contractAddress'],
            'amount': amount,
            'fee': float(tx.get('gasUsed', 0)) * float(tx.get('gasPrice', 0)) * self.coef,
            'gas': {
                'gas': float(tx['gas']),
                'gas_price': float(tx['gasPrice']) if tx.get('gasPrice') else None,
                'cumulative_gas_used': float(tx['cumulativeGasUsed']) if tx.get('cumulativeGasUsed') else None,
                'gas_used': float(tx['gasUsed']) if tx.get('gasUsed') else None
            },
            'hash': tx['hash'],
            'confirmations': int(tx['confirmations']) if tx.get('confirmations') else None,
            'confirmed': None,
            'is_error': tx.get('isError') == '1',
            'type': tx_type,
            'kind': 'transaction',
            'direction': direction,
            'token_data': token_data,
            'raw': tx
        }



from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class BtcAPI(BlockchainAPI):
    """
    coins: bitcoin-cash
    API docs: https://bch.btc.com/api-doc#API
    Explorer: https://btc.com
    """

    active = True

    symbol = 'BCH'
    base_url = 'https://bch-chain.api.btc.com/v3'
    rate_limit = 0
    coef = 1e-8
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/address/{address}',
    }

    @on_failure_return_none()
    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if response is None:
            return None

        if response['data'] is None:
            return None

        try:
            retval = response['data']['balance'] * self.coef
        except KeyError:
            return None

        return [{'symbol': self.symbol, 'amount': retval}]



from datetime import datetime

import pytz

from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class TrezorAPI(BlockchainAPI):
    """
    coins: bitcoin, litecoin
    API docs: https://github.com/trezor/blockbook/blob/master/docs/api.md
    Explorer:
    """

    active = True

    rate_limit = 0
    coef = 1e-8
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None
    xpub_support = True

    supported_requests = {
        'get_balance': '/api/v2/address/{address}',
        'get_balance_xpub': '/api/v2/xpub/{address}',
        'get_txs': '/api/v2/utxo/{address}?confirmed={confirmed}',
        'get_tx': '/api/v2/tx/{tx_hash}',
    }

    @on_failure_return_none()
    def get_balance(self):
        if len(self.address) == 111:
            response = self.request('get_balance_xpub',
                                    address=self.address)
        else:
            response = self.request('get_balance',
                                    address=self.address)

        if not response:
            return None

        retval = float(response.get('balance')) * self.coef
        return [{'symbol': self.symbol, 'amount': retval}]

    def get_txs(self, offset=None, limit=None, unconfirmed=False):
        response = self.request('get_txs',
                                address=self.address,
                                confirmed=not (unconfirmed))

        return [self.parse_tx(tx) for tx in response]

    def parse_tx(self, tx):
        txdata = self.request('get_tx',
                              tx_hash=tx['txid'])

        if self.address in txdata['vin'][0]['addresses']:
            direction = 'outgoing'
        else:
            direction = 'incoming'

        return {
            'date': datetime.fromtimestamp(txdata['blockTime'], pytz.utc),
            'from_address': txdata['vin'][0]['addresses'],
            'to_address': txdata['vout'][0]['addresses'],
            'amount': float(txdata['value']) * self.coef,
            'fee': float(txdata['fees']) * self.coef,
            'hash': tx['txid'],
            'confirmed': txdata['confirmations'],
            'is_error': False,
            'type': 'normal',
            'kind': 'transaction',
            'direction': direction,
            'status': 'confirmed' if txdata['confirmations'] > 0 else 'unconfirmed',
            'raw': txdata
        }


class Btc1TrezorAPI(TrezorAPI):
    base_url = 'https://btc1.trezor.io'
    symbol = 'BTC'


class Btc2TrezorAPI(TrezorAPI):
    base_url = 'https://btc2.trezor.io'
    symbol = 'BTC'


class Ltc1TrezorAPI(TrezorAPI):
    base_url = 'https://ltc1.trezor.io'
    symbol = 'LTC'



from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class TronscanAPI(BlockchainAPI):
    """
    coins: tron
    API docs: https://github.com/tronscan/tronscan-frontend/blob/dev2019/document/api.md
    Explorer: https://tronscan.org
    """

    active = True

    symbol = 'TRX'
    base_url = 'https://apilist.tronscan.org/api'
    rate_limit = 0
    coef = 1e-6
    max_items_per_page = None
    page_offset_step = None
    confirmed_num = None

    supported_requests = {
        'get_balance': '/account?address={address}',
        'get_trc10_tokenlist': '/token?limit=10000',
    }

    @on_failure_return_none()
    def get_balance(self):
        response = self.request('get_balance',
                                address=self.address)
        if not response:
            return None

        token_map = {}

        trc10_tokenlist = self.request('get_trc10_tokenlist')['data']
        for token in trc10_tokenlist:
            token_map[token['tokenID']] = {'abbr': token['abbr'], 'precision': token['precision'] }

        balances = []

        for coin in response['tokenBalances']:
            if coin['name'] == '_':
                symbol = self.symbol
                owner_address = self.address
                coin_coef = 1e-6
            else:
                symbol = token_map[int(coin['name'])]['abbr']
                owner_address = coin['owner_address']
                coin_coef = 1/pow(10,int(token_map[int(coin['name'])]['precision']))

            balances.append({'symbol': symbol,
                             'amount': float(coin['balance']) * coin_coef,
                             'address': owner_address})

        for coin in response['trc20token_balances']:
            balances.append({'symbol': coin['symbol'], 
                             'amount': float(coin['balance']) * (1/pow(10,int(coin['decimals']))),
                             'address': None })

        return balances



from blockapi.services import (
    BlockchainAPI,
    on_failure_return_none
)


class EthplorerAPI(BlockchainAPI):
    """
    Ethereum
    API docs: https://github.com/EverexIO/Ethplorer/wiki/Ethplorer-API
    Explorer: https://ethplorer.io
    """

    symbol = 'ETH'
    base_url = 'http://api.ethplorer.io'
    default_api_key = 'freekey'
    rate_limit = 0.5
    coef = 1e-18
    start_offset = None
    max_items_per_page = None
    page_offset_step = None

    supported_requests = {
        'get_info': '/getAddressInfo/{address}?apiKey={api_key}'
    }

    def __init__(self, address, api_key=None):
        if not api_key:
            api_key = self.default_api_key

        super().__init__(address, api_key)

        # cache info
        self._info = None

        if self.api_key != self.default_api_key:
            self.rate_limit = 0.1

    @on_failure_return_none()
    def get_balance(self):
        balances = []
        response = self.info

        if response.get('ETH', {}).get('balance'):
            balances.append({
                'symbol': self.symbol,
                'amount': response['ETH']['balance']
            })

        for token in response.get('tokens', []):
            info = token['tokenInfo']
            balances.append({
                'symbol': info['symbol'].upper(),
                'address': info['address'],
                'amount': token['balance'] * pow(10, -int(info['decimals'])),
                'name': info['name']
            })

        if balances == []:
            return None

        return balances

    @property
    def info(self):
        if not self._info:
            self._info = self.request(
                'get_info',
                address=self.address,
                api_key=self.api_key
            )
        return self._info
